#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from datetime import datetime, timedelta
from typing import NewType, Optional, Tuple, Iterable, List, Union
import itertools

from bbbmeetings.helpers import timestamp_to_datetime, seconds_to_timedelta
from bbbmeetings.connection import get_meetings
from bbbmeetings.types import *



class BBBServers():
    """
    Multiple servers represented as one
    """
    def __init__(self, servers: List['BBBServer']):
        self.servers = servers

    @classmethod
    def from_list(cls, servers: List['BBBServer']) -> 'Self':
        if not isinstance(servers, list):
            servers = list(servers)
        return cls(servers)

    def add(cls, servers: Union['BBBServer', List['BBBServer']]) -> 'Self':
        if not isinstance(servers, list):
            servers = list(servers)
        for server in servers:
            self.servers.append(server)
        return self

    def update_meetings(self) -> 'Self':
        """
        Update the meeting details by sendign a request to the BBBServers BBB-API
        """
        for server in self.servers:
            server.update_meetings()
        return self

    @property
    def meetings(self) -> List['Meeting']:
        """
        Return a flat list of all meetings on all servers
        """
        if any([s.last_update is None for s in self.servers]):
            for server in self.servers:
                server.update_meetings()

        return list(itertools.chain.from_iterable([s.meetings for s in self.servers]))

    @property
    def people(self) -> int:
        """
        Returns the count of participants of a meeting
        """
        return sum([m.participantCount for m in self.meetings])

    @property
    def biggest_meeting(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.people)

    @property
    def smallest_meeting(self) -> Optional['Meeting']:
        return min(self.meetings, key=lambda m: m.people)

    @property
    def n_meetings(self) -> int:
        return len(self.meetings)

    @property
    def listeners(self) -> int:
        return sum([m.listenerCount for m in self.meetings])

    @property
    def connected_with_mic(self) -> int:
        return sum([m.voiceParticipantCount for m in self.meetings])

    @property
    def video_active(self) -> int:
        return sum([m.videoCount for m in self.meetings])

    @property
    def moderators(self) -> int:
        return sum([m.moderatorCount for m in self.meetings])

    @property
    def longest_duration(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.duration)

    @property
    def most_listeners(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.listeners)

    @property
    def most_video_active(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.video_active)

    @property
    def most_moderators(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.moderators)

    @property
    def attendees(self) -> List['Attendees']:
        return list(itertools.chain.from_iterable([m.attendees for m in self.meetings]))




class BBBServer():
    """
    Represents a single Endpoint of the BBB-API (with a host address and a secret)
    The hostname could be something like "https://bbb.example.org/" while the secret
    is bascially a random string
    """
    def __init__(self, host, secret):
        self.host = host
        self.secret = secret
        self._meetings = []
        self.last_update = None

    @classmethod
    def from_dict(cls, d: dict):
        """
        Allows things like:
        ```
        s = BBBServer.from_dict({"host":"https://bbb.example.org", "secret":"12345"})
        ```
        """
        host   = d["host"]
        secret = d["secret"]
        return cls(host, secret)

    def update_meetings(self) -> 'Self':
        """
        Update the meeting details by sendign a request to the BBBServers BBB-API
        """
        self._meetings = get_meetings(self.host, self.secret)
        self.last_update = datetime.now()
        return self

    @property
    def meetings(self) -> List['Meeting']:
        if self.last_update is None:
            self.update_meetings()
        return self._meetings

    @property
    def people(self) -> int:
        """
        Returns the count of participants of a meeting
        """
        return sum([m.people for m in self.meetings])

    @property
    def biggest_meeting(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.people)

    @property
    def smallest_meeting(self) -> Optional['Meeting']:
        return min(self.meetings, key=lambda m: m.people)

    @property
    def n_meetings(self) -> int:
        return len(self.meetings)

    @property
    def listeners(self) -> int:
        return sum([m.listeners for m in self.meetings])

    @property
    def connected_with_mic(self) -> int:
        return sum([m.connected_with_mic for m in self.meetings])

    @property
    def video_active(self) -> int:
        return sum([m.cideo_active for m in self.meetings])

    @property
    def moderators(self) -> int:
        return sum([m.moderators for m in self.meetings])

    @property
    def longest_duration(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.duration)

    @property
    def most_listeners(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.listeners)

    @property
    def most_video_active(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.video_active)

    @property
    def most_moderators(self) -> Optional['Meeting']:
        return max(self.meetings, key=lambda m: m.moderators)

    @property
    def attendees(self) -> List['Attendees']:
        return list(itertools.chain.from_iterable([m.attendees for m in self.meetings]))





class Meeting():
    """
    A Meeting represents all the data the BBB-API returns about a Meeting
    """
    def __init__(self, name, meetingID, internalMeetingID, createTime, voiceBridge, dialNumber, attendeePW, moderatorPW, running, hasUserJoined, recording, hasBeenForciblyEnded, startTime, endTime, participantCount, listenerCount, voiceParticipantCount, videoCount, maxUsers, moderatorCount, attendees, isBreakout, parentRoom, breakoutRooms):
        self.name              = name
        self.meetingID         = meetingID
        self.internalMeetingID = internalMeetingID
        self.createTime        = createTime
        self.voiceBridge       = voiceBridge
        self.dialNumber        = dialNumber
        self.attendeePW        = attendeePW
        self.moderatorPW       = moderatorPW
        self.running           = running
        self.hasUserJoined     = hasUserJoined
        self.recording         = recording
        self.hasBeenForciblyEnded = hasBeenForciblyEnded
        self.startTime         = startTime
        self.endTime           = endTime
        self.participantCount  = participantCount
        self.listenerCount     = listenerCount
        self.voiceParticipantCount = voiceParticipantCount
        self.videoCount        = videoCount
        self.maxUsers          = maxUsers
        self.moderatorCount    = moderatorCount
        self.attendees         = attendees
        self.isBreakout        = isBreakout
        self.parentRoom        = parentRoom
        self.breakoutRooms     = breakoutRooms

    @classmethod
    def from_dict(cls, d: dict):
        name                   = d["meetingName"]
        meetingID              = d["meetingID"]
        internalMeetingID      = d["internalMeetingID"]
        createTime             = timestamp_to_datetime(d["createTime"])
        voiceBridge            = d["voiceBridge"]
        dialNumber             = d["dialNumber"]
        attendeePW             = d["attendeePW"]
        moderatorPW            = d["moderatorPW"]
        running                = d["running"] == "true"
        hasUserJoined          = d["hasUserJoined"] == "true"
        recording              = d["recording"] == "true"
        hasBeenForciblyEnded   = d["hasBeenForciblyEnded"] == "true"
        startTime              = timestamp_to_datetime(d["startTime"])
        endTime                = timestamp_to_datetime(d["endTime"])
        participantCount       = int(d["participantCount"])
        listenerCount          = int(d["listenerCount"])
        voiceParticipantCount  = int(d["voiceParticipantCount"])
        videoCount             = int(d["videoCount"])
        maxUsers               = int(d["maxUsers"])
        moderatorCount         = int(d["moderatorCount"])
        attendees              = []
        if d["attendees"] is not None:
            for x in d["attendees"].values():
                if isinstance(x, list):
                    for i in x:
                        attendees.append(Attendee.from_dict(i))
                else:
                    attendees.append(Attendee.from_dict(x))

        isBreakout             = d["isBreakout"] == "true"
        parentRoom = None
        breakoutRooms = []
        if isBreakout:
            parentRoom = d["parentMeetingID"]
        elif "breakoutRooms" in d.keys():
            for x in d["breakoutRooms"].values():
                if isinstance(x, list):
                    for i in x:
                        breakoutRooms.append(i)
                else:
                    breakoutRooms.append(x)

            
        return cls(name, meetingID, internalMeetingID, createTime, voiceBridge, dialNumber, attendeePW, moderatorPW, running, hasUserJoined, recording, hasBeenForciblyEnded, startTime, endTime, participantCount, listenerCount, voiceParticipantCount, videoCount, maxUsers, moderatorCount, attendees, isBreakout, parentRoom, breakoutRooms)

    def __str__(self):
        s = f"""Meeting
    name:                  {self.name}
    meetingID:             {self.meetingID}
    internalMeetingID:     {self.internalMeetingID}
    createTime:            {self.createTime}
    voiceBridge:           {self.voiceBridge}
    dialNumber:            {self.dialNumber}
    attendeePW:            {self.attendeePW}
    moderatorPW:           {self.moderatorPW}
    running:               {self.running}
    duration:              {self.duration}
    hasUserJoined:         {self.hasUserJoined}
    recording:             {self.recording}
    hasBeenForciblyEnded:  {self.hasBeenForciblyEnded}
    startTime:             {self.startTime}
    endTime:               {self.endTime}
    participantCount:      {self.participantCount}
    listenerCount:         {self.listenerCount}
    voiceParticipantCount: {self.voiceParticipantCount}
    videoCount:            {self.videoCount}
    maxUsers:              {self.maxUsers}
    moderatorCount:        {self.moderatorCount}
    attendees:             {', '.join([a.fullName for a in self.attendees])}
    isBreakout:            {self.isBreakout}
    parentRoom:            {self.parentRoom}
    breakoutRooms:         {', '.join([x for x in self.breakoutRooms])}
    """
        return s

    def __repr__(self):
        return str(self)

    @property
    def duration(self) -> timedelta:
        return datetime.now() - self.startTime

    @property
    def people(self) -> int:
        return self.participantCount

    @property
    def listeners(self) -> int:
        return self.listenerCount

    @property
    def connected_with_mic(self) -> int:
        return self.voiceParticipantCount

    @property
    def video_active(self) -> int:
        return  self.videoCount

    @property
    def moderators(self) -> int:
        return self.moderatorCount



class Attendee():
    """
    An Attendee is someone who is joined to a Meeting, as represented by the BBB-API
    """
    def __init__(self, userID, fullName, role, isPresenter, isListeningOnly, hasJoinedVoice, hasVideo, clientType):
        self.userID          = userID
        self.fullName        = fullName
        self.role            = role
        self.isPresenter     = isPresenter
        self.isListeningOnly = isListeningOnly
        self.hasJoinedVoice  = hasJoinedVoice
        self.hasVideo        = hasVideo
        self.clientType      = clientType

    @classmethod
    def from_dict(cls, d: dict):
        userID           = d["userID"]
        fullName         = d["fullName"]
        role             = d["role"]
        isPresenter      = d["isPresenter"] == "true"
        isListeningOnly  = d["isListeningOnly"] == "true"
        hasJoinedVoice   = d["hasJoinedVoice"] == "true"
        hasVideo         = d["hasVideo"] == "true"
        clientType       = d["clientType"]

        return cls(userID, fullName, role, isPresenter, isListeningOnly, hasJoinedVoice, hasVideo, clientType)

    @property
    def name(self) -> Optional[str]:
        return self.fullName



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from datetime import datetime
from typing import NewType, Optional, Tuple, Iterable, List
import hashlib
import requests
import xmltodict

from bbbmeetings.types import *
import bbbmeetings as classes

def generate_checksum(call_name: str, query_string: str, secret: Secret) -> str:
    """
    Generate Checksum for the request header (passed as value for `?checksum=`)
    """
    m = hashlib.sha1()
    m.update(call_name.encode('utf-8'))
    m.update(query_string.encode('utf-8'))
    m.update(secret.encode('utf-8'))
    return m.hexdigest()


def send_request(bbb_url: Url, secret: Secret, call_name: str, query_string: str = "") -> Optional[CheckedResponse]:
    """
    Requests a certain call and query from the BBB-API. Returns None if anything fails or a checked OrderedDict otherwise
    """
    bbb_url = bbb_url.rstrip("/")
    checksum = generate_checksum(call_name, query_string, secret)
    url = "{}/bigbluebutton/api/{}?checksum={}".format(bbb_url, call_name, checksum)

    try:
        r = requests.get(url, timeout=3)
        if r.ok:
            d = xmltodict.parse(r.text)
            checked = clean_and_check_dict(d)
            return checked
        else:
            print(f"Request to Server failed with {r.status_code}")
            return None
    except requests.exceptions.Timeout as e:
        print(f"Request to Server timed out")
        # Offline Server!
        return None
    except:
        # Offline Server!
        return None


def clean_and_check_dict(d: UnparsedResponse) -> Optional[CheckedResponse]:
    """
    Checks if the response was a SUCCESS and returns the unpacked OrderedDict, None otherwise
    """
    if not "response" in d.keys():
        return None
    elif d["response"]["returncode"].upper() != "SUCCESS":
        return None
    return d["response"]


def get_meetings(bbb_url: Url, secret: Secret) -> List['Meeting']:
    checked_dict = send_request(bbb_url, secret, "getMeetings")
    if checked_dict is None:
        return []
    elif not "meetings" in checked_dict or checked_dict["meetings"] is None:
        return []
    else:
        meetings = []
        for d in checked_dict["meetings"].values():
            if isinstance(d, list):
                for x in d:
                    meetings.append(classes.Meeting.from_dict(x))
            else:
                meetings.append(classes.Meeting.from_dict(d))
        return meetings


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import NewType, Optional, Tuple, Iterable, List
from collections import OrderedDict

Secret = NewType('Secret', str)
Url    = NewType('Url', str)
UnparsedResponse    = NewType('UnparsedResponse', OrderedDict)
CheckedResponse    = NewType('CheckedResponse', OrderedDict)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import NewType, Optional, Tuple, Iterable, List
from datetime import datetime, timedelta

def timestamp_to_datetime(timestamp: int) -> Optional[datetime]:
    if int(timestamp) == 0:
        return None
    # BBB Timestamps are too precise, remove last three digits
    timestamp = int(timestamp[:-3])
    return datetime.fromtimestamp(timestamp)


def seconds_to_timedelta(seconds: int) -> Optional[timedelta]:
    if int(seconds) == 0:
        return None
    return timedelta(seconds=int(seconds))


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bbbmeetings']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.1,<3.0.0', 'xmltodict>=0.12.0,<0.13.0']

setup_kwargs = {
    'name': 'bbbmeetings',
    'version': '0.1.2',
    'description': 'A module for reading bbb meetings from bbb servers',
    'long_description': '# bbbmeetings\n\nA Library that helps with reading information regarding Big Blue Button via the BBB-API. The goal of bbbmeetings is to make it easy to generate stats about one or more BBB-Instances.\n\nIt was a non-goal to create a full coverage of the BBB-API, for ending meetings, creating meetings, managing recordings etc.\n\n\n\n## Installation\n\nVia pip\n\n```\npip install bbbmeetings\n```\n\n\n\n## Usage\n\nFirst we need to create a BBB Server. To allow the connection with that server you need to know its secret, which can be found using `bbb-conf -- secret` as [mentioned here](https://docs.bigbluebutton.org/admin/bbb-conf.html#--secret). You can create a server like this:\n\n```python\nfrom bbbmeetings import BBBServer\nserver = BBBServer(host="https://bbb.example.org", secret="!#%Random!#$%Fsecret!#$%")\n```\n\nFor convenience bbbmeetings offers a way of treating multiple servers as one server:\n\n```python\nfrom bbbmeetings import BBBServers, BBBServer\n\n# Create a BBBServers Object from multiple BBBServer-Objects\nservers = BBBServers.from_list([\n        BBBServer("https://bbb1.example.org", "!#%Random!#$%Fsecret!#$%"),\n        BBBServer("https://bbb2.example.org", "!#%Random!#$%Fsecret!#$%"),\n        BBBServer("https://bbb3.example.org", "!#%Random!#$%Fsecret!#$%"),\n        BBBServer("https://bbb4.example.org", "!#%Random!#$%Fsecret!#$%"),\n        BBBServer("https://bbb5.example.org", "!#%Random!#$%Fsecret!#$%"),\n    ])\n```\n\nMost methods on `BBBServer` and `BBBServers` are the same, but in the following examples I will use `servers` to showcase how this module may be used:\n\n```python\n# Get a list of meetings running on all servers and print all details about them\nfor meeting in servers.meetings:\n    print(meeting)\n```\n\nYou can also get some numbers directly (in this case I use python\'s formatting strings):\n\n```python\nprint(f"On all servers there are {servers.people} people in {servers.n_meetings} meetings. {servers.video_active} people have their webcam on, {servers.connected_with_mic} connected with mic. There are {servers.moderators} moderators.")\n```\n\nIf you want to have a list of all people in all meetings on all servers run:\n\n```python\nprint(f"The names of everyone on all servers are {\', \'.join([a.name for a in servers.attendees])}")\n```\n\nIf you just want a list with the names of Presenters you can run:\n\n```python\npresenter_names = [a.name for a in servers.attendees if a.isPresenter]\n```\n\nOr if you want to know the longest running meeting:\n\n```python\nprint(f"Longest running meeting is {servers.longest_duration.name} with a duration of {servers.longest_duration.duration}")\n```\n\n',
    'author': 'David Huss',
    'author_email': 'david.huss@hfbk-hamburg.de',
    'maintainer': 'David Huss',
    'maintainer_email': 'david.huss@hfbk-hamburg.de',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)
